@@ -7,14 +7,12 @@
 from __future__ import absolute_import, print_function
 
 import jupyter_server
-import asyncio
 import binascii
 import datetime
 import errno
 import gettext
 import hashlib
 import hmac
-import importlib
 import io
 import ipaddress
 import json
@@ -33,25 +31,32 @@ import time
 import warnings
 import webbrowser
 import urllib
+import inspect
+import pathlib
 
-from types import ModuleType
 from base64 import encodebytes
+try:
+    import resource
+except ImportError:
+    # Windows
+    resource = None
+
 from jinja2 import Environment, FileSystemLoader
 
+from jupyter_core.paths import secure_write
 from jupyter_server.transutils import trans, _
-from jupyter_server.utils import secure_write, run_sync
+from jupyter_server.utils import run_sync
+
+# the minimum viable tornado version: needs to be kept in sync with setup.py
+MIN_TORNADO = (6, 1, 0)
 
-# check for tornado 3.1.0
 try:
     import tornado
-except ImportError as e:
-    raise ImportError(_("The Jupyter Server requires tornado >= 4.0")) from e
-try:
-    version_info = tornado.version_info
-except AttributeError as e:
-    raise ImportError(_("The Jupyter Server requires tornado >= 4.0, but you have < 1.1.0")) from e
-if version_info < (4,0):
-    raise ImportError(_("The Jupyter Server requires tornado >= 4.0, but you have %s") % tornado.version)
+    assert tornado.version_info >= MIN_TORNADO
+except (ImportError, AttributeError, AssertionError) as e:  # pragma: no cover
+    raise ImportError(
+        _("The Jupyter Server requires tornado >=%s.%s.%s") % MIN_TORNADO
+    ) from e
 
 from tornado import httpserver
 from tornado import ioloop
@@ -69,8 +74,8 @@ from .base.handlers import MainHandler, RedirectWithParams, Template404
 from .log import log_request
 from .services.kernels.kernelmanager import MappingKernelManager, AsyncMappingKernelManager
 from .services.config import ConfigManager
-from .services.contents.manager import ContentsManager
-from .services.contents.filemanager import FileContentsManager
+from .services.contents.manager import AsyncContentsManager, ContentsManager
+from .services.contents.filemanager import AsyncFileContentsManager, FileContentsManager
 from .services.contents.largefilemanager import LargeFileManager
 from .services.sessions.sessionmanager import SessionManager
 from .gateway.managers import GatewayKernelManager, GatewayKernelSpecManager, GatewaySessionManager, GatewayClient
@@ -98,11 +103,18 @@ from jupyter_core.paths import jupyter_runtime_dir, jupyter_path
 from jupyter_server._sysinfo import get_sys_info
 
 from ._tz import utcnow, utcfromtimestamp
-from .utils import url_path_join, check_pid, url_escape, urljoin, pathname2url
+from .utils import (
+    url_path_join,
+    check_pid,
+    url_escape,
+    urljoin,
+    pathname2url
+)
 
 from jupyter_server.extension.serverextension import ServerExtensionApp
 from jupyter_server.extension.manager import ExtensionManager
 from jupyter_server.extension.config import ExtensionConfigManager
+from jupyter_server.traittypes import TypeFromClasses
 
 #-----------------------------------------------------------------------------
 # Module globals
@@ -247,6 +259,7 @@ class ServerWebApplication(web.Application):
             disable_check_xsrf=jupyter_app.disable_check_xsrf,
             allow_remote_access=jupyter_app.allow_remote_access,
             local_hostnames=jupyter_app.local_hostnames,
+            authenticate_prometheus=jupyter_app.authenticate_prometheus,
 
             # managers
             kernel_manager=kernel_manager,
@@ -511,9 +524,28 @@ flags['allow-root']=(
     _("Allow the server to be run from root user.")
 )
 flags["no-browser"] = (
-    {"ServerApp": {"open_browser": False}},
+    {
+        "ServerApp": {"open_browser": False},
+        "ExtensionApp": {"open_browser": False}
+    },
     _("Prevent the opening of the default url in the browser."),
 )
+flags["debug"] = (
+    {
+        'ServerApp': {'log_level': 'DEBUG'},
+        'ExtensionApp': {'log_level': 'DEBUG'}
+    },
+    _("Set debug level for the extension and underlying server applications.")
+)
+flags['autoreload'] = (
+    {'ServerApp': {'autoreload': True}},
+    """Autoreload the webapp
+    Enable reloading of the tornado webapp and all imported Python packages
+    when any changes are made to any Python src files in server or
+    extensions.
+    """
+)
+
 
 # Add notebook manager flags
 flags.update(boolean_flag('script', 'FileContentsManager.save_script',
@@ -553,8 +585,8 @@ class ServerApp(JupyterApp):
     aliases = Dict(aliases)
 
     classes = [
-            KernelManager, Session, MappingKernelManager, KernelSpecManager,
-            ContentsManager, FileContentsManager, NotebookNotary,
+            KernelManager, Session, MappingKernelManager, KernelSpecManager, AsyncMappingKernelManager,
+            ContentsManager, FileContentsManager, AsyncContentsManager, AsyncFileContentsManager, NotebookNotary,
             GatewayKernelManager, GatewayKernelSpecManager, GatewaySessionManager, GatewayClient
         ]
 
@@ -595,10 +627,15 @@ class ServerApp(JupyterApp):
         return u"%(color)s[%(levelname)1.1s %(asctime)s.%(msecs).03d %(name)s]%(end_color)s %(message)s"
 
     # file to be opened in the Jupyter server
-    file_to_run = Unicode('', config=True)
+    file_to_run = Unicode('',
+        help="Open the named file when the application is launched."
+    ).tag(config=True)
+
+    file_url_prefix = Unicode('notebooks',
+        help="The URL prefix where files are opened directly."
+    ).tag(config=True)
 
     # Network related information
-
     allow_origin = Unicode('', config=True,
         help="""Set the Access-Control-Allow-Origin header
 
@@ -629,6 +666,10 @@ class ServerApp(JupyterApp):
         help=_("Whether to allow the user to run the server as root.")
     )
 
+    autoreload = Bool(False, config=True,
+        help= ("Reload the webapp when changes are made to any Python src files.")
+    )
+
     default_url = Unicode('/', config=True,
         help=_("The default URL to redirect to from `/`")
     )
@@ -699,7 +740,7 @@ class ServerApp(JupyterApp):
 
     @default('cookie_secret_file')
     def _default_cookie_secret_file(self):
-        return os.path.join(self.runtime_dir, 'jupytr_cookie_secret')
+        return os.path.join(self.runtime_dir, 'jupyter_cookie_secret')
 
     cookie_secret = Bytes(b'', config=True,
         help="""The random bytes used to secure cookies.
@@ -758,6 +799,30 @@ class ServerApp(JupyterApp):
             self._token_generated = True
             return binascii.hexlify(os.urandom(24)).decode('ascii')
 
+    min_open_files_limit = Integer(config=True,
+        help="""
+        Gets or sets a lower bound on the open file handles process resource
+        limit. This may need to be increased if you run into an
+        OSError: [Errno 24] Too many open files.
+        This is not applicable when running on Windows.
+        """)
+
+    @default('min_open_files_limit')
+    def _default_min_open_files_limit(self):
+        if resource is None:
+            # Ignoring min_open_files_limit because the limit cannot be adjusted (for example, on Windows)
+            return None
+
+        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
+
+        DEFAULT_SOFT = 4096
+        if hard >= DEFAULT_SOFT:
+            return DEFAULT_SOFT
+
+        self.log.debug("Default value for min_open_files_limit is ignored (hard=%r, soft=%r)", hard, soft)
+
+        return soft
+
     max_body_size = Integer(512 * 1024 * 1024, config=True,
         help="""
         Sets the maximum allowed size of the client request body, specified in
@@ -879,6 +944,22 @@ class ServerApp(JupyterApp):
         else:
             return not addr.is_loopback
 
+    use_redirect_file = Bool(True, config=True,
+        help="""Disable launching browser by redirect file
+     For versions of notebook > 5.7.2, a security feature measure was added that
+     prevented the authentication token used to launch the browser from being visible.
+     This feature makes it difficult for other users on a multi-user system from
+     running code in your Jupyter session as you.
+     However, some environments (like Windows Subsystem for Linux (WSL) and Chromebooks),
+     launching a browser using a redirect file can lead the browser failing to load.
+     This is because of the difference in file structures/paths between the runtime and
+     the browser.
+
+     Disabling this setting to False will disable this behavior, allowing the browser
+     to launch by using a URL and visible token (as before).
+     """
+    )
+
     local_hostnames = List(Unicode(), ['localhost'], config=True,
        help="""Hostnames to allow as local when allow_remote_access is False.
 
@@ -1023,15 +1104,46 @@ class ServerApp(JupyterApp):
         help="""If True, display controls to shut down the Jupyter server, such as menu items or buttons."""
     )
 
-    contents_manager_class = Type(
+    # REMOVE in VERSION 2.0
+    # Temporarily allow content managers to inherit from the 'notebook'
+    # package. We will deprecate this in the next major release.
+    contents_manager_class = TypeFromClasses(
         default_value=LargeFileManager,
-        klass=ContentsManager,
+        klasses=[
+            'jupyter_server.services.contents.manager.ContentsManager',
+            'notebook.services.contents.manager.ContentsManager'
+        ],
         config=True,
         help=_('The content manager class to use.')
     )
 
+    # Throws a deprecation warning to notebook based contents managers.
+    @observe('contents_manager_class')
+    def _observe_contents_manager_class(self, change):
+        new = change['new']
+        # If 'new' is a class, get a string representing the import
+        # module path.
+        if inspect.isclass(new):
+            new = new.__module__
+
+        if new.startswith('notebook'):
+            self.log.warning(
+                "The specified 'contents_manager_class' class inherits a manager from the "
+                "'notebook' package. This is not guaranteed to work in future "
+                "releases of Jupyter Server. Instead, consider switching the "
+                "manager to inherit from the 'jupyter_server' managers. "
+                "Jupyter Server will temporarily allow 'notebook' managers "
+                "until its next major release (2.x)."
+            )
+
+    @observe('notebook_dir')
+    def _update_notebook_dir(self, change):
+        self.log.warning(_("notebook_dir is deprecated, use root_dir"))
+        self.root_dir = change['new']
+
     kernel_manager_class = Type(
-        default_value=MappingKernelManager,
+        default_value=AsyncMappingKernelManager,
+        klass=MappingKernelManager,
         config=True,
         help=_('The kernel manager class to use.')
     )
@@ -1095,6 +1207,13 @@ class ServerApp(JupyterApp):
         basename = "jpserver-%s-open.html" % os.getpid()
         return os.path.join(self.runtime_dir, basename)
 
+    browser_open_file_to_run = Unicode()
+
+    @default('browser_open_file_to_run')
+    def _default_browser_open_file_to_run(self):
+        basename = "jpserver-file-to-run-%s-open.html" % os.getpid()
+        return os.path.join(self.runtime_dir, basename)
+
     pylab = Unicode('disabled', config=True,
         help=_("""
         DISABLED: use %pylab or %matplotlib in the notebook to enable matplotlib.
@@ -1121,16 +1240,22 @@ class ServerApp(JupyterApp):
 
     @observe('notebook_dir')
     def _update_notebook_dir(self, change):
+        if self._root_dir_set:
+            # only use deprecated config if new config is not set
+            return
         self.log.warning(_("notebook_dir is deprecated, use root_dir"))
         self.root_dir = change['new']
 
-    root_dir = Unicode(config=True,
+    root_dir = Unicode(
+        config=True,
         help=_("The directory to use for notebooks and kernels.")
     )
+    _root_dir_set = False
 
     @default('root_dir')
     def _default_root_dir(self):
         if self.file_to_run:
+            self._root_dir_set = True
             return os.path.dirname(os.path.abspath(self.file_to_run))
         else:
             return py3compat.getcwd()
@@ -1148,9 +1273,13 @@ class ServerApp(JupyterApp):
             # If we receive a non-absolute path, make it absolute.
             value = os.path.abspath(value)
         if not os.path.isdir(value):
-            raise TraitError(trans.gettext("No such notebook dir: '%r'") % value)
+            raise TraitError(trans.gettext("No such directory: '%r'") % value)
         return value
 
+    @observe('root_dir')
+    def _root_dir_changed(self, change):
+        self._root_dir_set = True
+
     @observe('server_extensions')
     def _update_server_extensions(self, change):
         self.log.warning(_("server_extensions is deprecated, use jpserver_extensions"))
@@ -1200,6 +1329,25 @@ class ServerApp(JupyterApp):
          is not available.
          """))
 
+    authenticate_prometheus = Bool(
+        True,
+        help=""""
+        Require authentication to access prometheus metrics.
+        """,
+        config=True
+    )
+
+    _starter_app = Instance(
+        default_value=None,
+        allow_none=True,
+        klass='jupyter_server.extension.application.ExtensionApp'
+    )
+
+    @property
+    def starter_app(self):
+        """Get the Extension that started this server."""
+        return self._starter_app
+
     def parse_command_line(self, argv=None):
 
         super(ServerApp, self).parse_command_line(argv)
@@ -1278,6 +1426,7 @@ class ServerApp(JupyterApp):
         if self.allow_origin_pat:
             self.tornado_settings['allow_origin_pat'] = re.compile(self.allow_origin_pat)
         self.tornado_settings['allow_credentials'] = self.allow_credentials
+        self.tornado_settings['autoreload'] = self.autoreload
         self.tornado_settings['cookie_options'] = self.cookie_options
         self.tornado_settings['get_secure_cookie_kwargs'] = self.get_secure_cookie_kwargs
         self.tornado_settings['token'] = self.token
@@ -1324,6 +1473,23 @@ class ServerApp(JupyterApp):
 
         self.login_handler_class.validate_security(self, ssl_options=self.ssl_options)
 
+    def init_resources(self):
+        """initialize system resources"""
+        if resource is None:
+            self.log.debug('Ignoring min_open_files_limit because the limit cannot be adjusted (for example, on Windows)')
+            return
+
+        old_soft, old_hard = resource.getrlimit(resource.RLIMIT_NOFILE)
+        soft = self.min_open_files_limit
+        hard = old_hard
+        if old_soft < soft:
+            if hard < soft:
+                hard = soft
+            self.log.debug(
+                'Raising open file limit: soft {}->{}; hard {}->{}'.format(old_soft, soft, old_hard, hard)
+            )
+            resource.setrlimit(resource.RLIMIT_NOFILE, (soft, hard))
+
     @property
     def display_url(self):
         if self.custom_display_url:
@@ -1604,34 +1770,15 @@ class ServerApp(JupyterApp):
 
     @staticmethod
     def _init_asyncio_patch():
-        """set default asyncio policy to be compatible with tornado
-        Tornado 6 (at least) is not compatible with the default
-        asyncio implementation on Windows
-        Pick the older SelectorEventLoopPolicy on Windows
-        if the known-incompatible default policy is in use.
-        do this as early as possible to make it a low priority and overrideable
-        ref: https://github.com/tornadoweb/tornado/issues/2608
-        FIXME: if/when tornado supports the defaults in asyncio,
-               remove and bump tornado requirement for py38
-        """
-        if sys.platform.startswith("win") and sys.version_info >= (3, 8):
-            import asyncio
-            try:
-                from asyncio import (
-                    WindowsProactorEventLoopPolicy,
-                    WindowsSelectorEventLoopPolicy,
-                )
-            except ImportError:
-                pass
-                # not affected
-            else:
-                if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
-                    # WindowsProactorEventLoopPolicy is not compatible with tornado 6
-                    # fallback to the pre-3.8 default of Selector
-                    asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())
+        """no longer needed with tornado 6.1"""
+        warnings.warn(
+            """ServerApp._init_asyncio_patch called, and is longer needed for """
+            """tornado 6.1+, and will be removed in a future release.""",
+            DeprecationWarning
+        )
 
     @catch_config_error
-    def initialize(self, argv=None, find_extensions=True, new_httpserver=True):
+    def initialize(self, argv=None, find_extensions=True, new_httpserver=True, starter_extension=None):
         """Initialize the Server application class, configurables, web application, and http server.
 
         Parameters
@@ -1647,12 +1794,14 @@ class ServerApp(JupyterApp):
         new_httpserver: bool
             If True, a tornado HTTPServer instance will be created and configured for the Server Web
             Application. This will set the http_server attribute of this class.
+
+        starter_extension: str
+            If given, it references the name of an extension point that started the Server.
+            We will try to load configuration from extension point
         """
-        self._init_asyncio_patch()
         # Parse command line, load ServerApp config files,
         # and update ServerApp config.
-        super(ServerApp, self).initialize(argv)
-        # Initialize all components of the ServerApp.
+        super(ServerApp, self).initialize(argv=argv)
         if self._dispatching:
             return
         # Then, use extensions' config loading mechanism to
@@ -1661,6 +1810,20 @@ class ServerApp(JupyterApp):
             self.find_server_extensions()
         self.init_logging()
         self.init_server_extensions()
+
+        # Special case the starter extension and load
+        # any server configuration is provides.
+        if starter_extension:
+            # Configure ServerApp based on named extension.
+            point = self.extension_manager.extension_points[starter_extension]
+            # Set starter_app property.
+            if point.app:
+                self._starter_app = point.app
+            # Load any configuration that comes from the Extension point.
+            self.update_config(Config(point.config))
+
+        # Initialize other pieces of the server.
+        self.init_resources()
         self.init_configurables()
         self.init_components()
         self.init_webapp()
@@ -1730,18 +1893,38 @@ class ServerApp(JupyterApp):
             os.unlink(self.info_file)
         except OSError as e:
             if e.errno != errno.ENOENT:
-                raise
+                raise;
 
-    def write_browser_open_file(self):
-        """Write an nbserver-<pid>-open.html file
-
-        This can be used to open the notebook in a browser
+    def _resolve_file_to_run_and_root_dir(self):
+        """Returns a relative path from file_to_run
+        to root_dir. If root_dir and file_to_run
+        are incompatible, i.e. on different subtrees,
+        crash the app and log a critical message. Note
+        that if root_dir is not configured and file_to_run
+        is configured, root_dir will be set to the parent
+        directory of file_to_run.
         """
-        # default_url contains base_url, but so does connection_url
-        open_url = self.default_url[len(self.base_url):]
+        rootdir_abspath = pathlib.Path(self.root_dir).resolve()
+        file_rawpath = pathlib.Path(self.file_to_run)
+        combined_path = (rootdir_abspath / file_rawpath).resolve()
+        is_child = str(combined_path).startswith(str(rootdir_abspath))
 
-        with open(self.browser_open_file, 'w', encoding='utf-8') as f:
-            self._write_browser_open_file(open_url, f)
+        if is_child:
+            if combined_path.parent != rootdir_abspath:
+                self.log.debug(
+                    "The `root_dir` trait is set to a directory that's not "
+                    "the immediate parent directory of `file_to_run`. Note that "
+                    "the server will start at `root_dir` and open the "
+                    "the file from the relative path to the `root_dir`."
+                )
+            return str(combined_path.relative_to(rootdir_abspath))
+
+        self.log.critical(
+            "`root_dir` and `file_to_run` are incompatible. They "
+            "don't share the same subtrees. Make sure `file_to_run` "
+            "is on the same path as `root_dir`."
+        )
+        self.exit(1)
 
     def _write_browser_open_file(self, url, fh):
         if self.token:
@@ -1752,8 +1935,53 @@ class ServerApp(JupyterApp):
         template = jinja2_env.get_template('browser-open.html')
         fh.write(template.render(open_url=url, base_url=self.base_url))
 
+    def write_browser_open_files(self):
+        """Write an `browser_open_file` and `browser_open_file_to_run` files
+
+        This can be used to open a file directly in a browser.
+        """
+        # default_url contains base_url, but so does connection_url
+        self.write_browser_open_file()
+
+        # Create a second browser open file if
+        # file_to_run is set.
+        if self.file_to_run:
+            # Make sure file_to_run and root_dir are compatible.
+            file_to_run_relpath = self._resolve_file_to_run_and_root_dir()
+
+            file_open_url = url_escape(
+                url_path_join(self.file_url_prefix, *file_to_run_relpath.split(os.sep))
+            )
+
+            with open(self.browser_open_file_to_run, 'w', encoding='utf-8') as f:
+                self._write_browser_open_file(file_open_url, f)
+
+    def write_browser_open_file(self):
+        """Write an jpserver-<pid>-open.html file
+
+        This can be used to open the notebook in a browser
+        """
+        # default_url contains base_url, but so does connection_url
+        open_url = self.default_url[len(self.base_url):]
+
+        with open(self.browser_open_file, 'w', encoding='utf-8') as f:
+            self._write_browser_open_file(open_url, f)
+
+    def remove_browser_open_files(self):
+        """Remove the `browser_open_file` and `browser_open_file_to_run` files
+        created for this server.
+
+        Ignores the error raised when the file has already been removed.
+        """
+        self.remove_browser_open_file()
+        try:
+            os.unlink(self.browser_open_file_to_run)
+        except OSError as e:
+            if e.errno != errno.ENOENT:
+                raises
+
     def remove_browser_open_file(self):
-        """Remove the nbserver-<pid>-open.html file created for this server.
+        """Remove the jpserver-<pid>-open.html file created for this server.
 
         Ignores the error raised when the file has already been removed.
         """
@@ -1763,6 +1991,28 @@ class ServerApp(JupyterApp):
             if e.errno != errno.ENOENT:
                 raise
 
+    def _prepare_browser_open(self):
+        if not self.use_redirect_file:
+            uri = self.default_url[len(self.base_url):]
+
+            if self.token:
+                uri = url_concat(uri, {'token': self.token})
+
+        if self.file_to_run:
+            # Create a separate, temporary open-browser-file
+            # pointing at a specific file.
+            open_file = self.browser_open_file_to_run
+        else:
+            # otherwise, just return the usual open browser file.
+            open_file = self.browser_open_file
+
+        if self.use_redirect_file:
+            assembled_url = urljoin('file:', pathname2url(open_file))
+        else:
+            assembled_url = url_path_join(self.connection_url, uri)
+
+        return assembled_url, open_file
+
     def launch_browser(self):
         try:
             browser = webbrowser.get(self.browser or None)
@@ -1773,24 +2023,9 @@ class ServerApp(JupyterApp):
         if not browser:
             return
 
-        if self.file_to_run:
-            if not os.path.exists(self.file_to_run):
-                self.log.critical(_("%s does not exist") % self.file_to_run)
-                self.exit(1)
+        assembled_url, _ = self._prepare_browser_open()
 
-            relpath = os.path.relpath(self.file_to_run, self.root_dir)
-            uri = url_escape(url_path_join('notebooks', *relpath.split(os.sep)))
-
-            # Write a temporary file to open in the browser
-            fd, open_file = tempfile.mkstemp(suffix='.html')
-            with open(fd, 'w', encoding='utf-8') as fh:
-                self._write_browser_open_file(uri, fh)
-        else:
-            open_file = self.browser_open_file
-
-        b = lambda: browser.open(
-            urljoin('file:', pathname2url(open_file)),
-            new=self.webbrowser_open_new)
+        b = lambda: browser.open(assembled_url, new=self.webbrowser_open_new)
         threading.Thread(target=b).start()
 
     def start_app(self):
@@ -1817,8 +2052,9 @@ class ServerApp(JupyterApp):
                  "resources section at https://jupyter.org/community.html."))
 
         self.write_server_info_file()
-        self.write_browser_open_file()
+        self.write_browser_open_files()
 
+        # Handle the browser opening.
         if self.open_browser:
             self.launch_browser()
 
@@ -1833,6 +2069,14 @@ class ServerApp(JupyterApp):
                 '    %s' % self.display_url,
             ]))
 
+    def _cleanup(self):
+        """General cleanup of files and kernels created
+        by this instance ServerApp.
+        """
+        self.remove_server_info_file()
+        self.remove_browser_open_files()
+        self.cleanup_kernels()
+
     def start_ioloop(self):
         """Start the IO Loop."""
         self.io_loop = ioloop.IOLoop.current()
@@ -1846,9 +2090,7 @@ class ServerApp(JupyterApp):
         except KeyboardInterrupt:
             self.log.info(_("Interrupted..."))
         finally:
-            self.remove_server_info_file()
-            self.remove_browser_open_file()
-            self.cleanup_kernels()
+            self._cleanup()
 
     def start(self):
         """ Start the Jupyter server app, after initialization
